@@ -22,13 +22,13 @@ import sys
 import time
 import uuid
 
-from positional import positional
 import requests
 import six
 from six.moves import urllib
 
 import keystoneauth1
 from keystoneauth1 import _utils as utils
+from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 
 try:
@@ -50,8 +50,6 @@ DEFAULT_USER_AGENT = 'keystoneauth1/%s %s %s/%s' % (
 # here and we'll add it to the list as required.
 _LOG_CONTENT_TYPES = set(['application/json'])
 
-_logger = utils.get_logger(__name__)
-
 
 def _construct_session(session_obj=None):
     # NOTE(morganfainberg): if the logic in this function changes be sure to
@@ -65,6 +63,22 @@ def _construct_session(session_obj=None):
     return session_obj
 
 
+def _mv_legacy_headers_for_service(mv_service_type):
+    """Workaround for services that predate standardization.
+
+    TODO(sdague): eventually convert this to using os-service-types
+    and put the logic there. However, right now this is so little
+    logic, inlining it for release is a better call.
+
+    """
+    headers = []
+    if mv_service_type == "compute":
+        headers.append("X-OpenStack-Nova-API-Version")
+    elif mv_service_type == "baremetal":
+        headers.append("X-OpenStack-Ironic-API-Version")
+    return headers
+
+
 class _JSONEncoder(json.JSONEncoder):
 
     def default(self, o):
@@ -221,6 +235,15 @@ class Session(object):
                                        will be added to the user agent. This
                                        can be used by libraries that are part
                                        of the communication process.
+    :param dict discovery_cache: A dict to be used for caching of discovery
+                                 information. This is normally managed
+                                 transparently, but if the user wants to
+                                 share a single cache across multiple sessions
+                                 that do not share an auth plugin, it can
+                                 be provided here. (optional, defaults to
+                                 None which means automatically manage)
+    :param bool split_loggers: Split the logging of requests across multiple
+                               loggers instead of just one. Defaults to False.
     """
 
     user_agent = None
@@ -229,11 +252,11 @@ class Session(object):
 
     _DEFAULT_REDIRECT_LIMIT = 30
 
-    @positional(2)
     def __init__(self, auth=None, session=None, original_ip=None, verify=True,
                  cert=None, timeout=None, user_agent=None,
                  redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
-                 app_name=None, app_version=None, additional_user_agent=None):
+                 app_name=None, app_version=None, additional_user_agent=None,
+                 discovery_cache=None, split_loggers=None):
 
         self.auth = auth
         self.session = _construct_session(session)
@@ -247,6 +270,12 @@ class Session(object):
         self.app_version = app_version
         self.additional_user_agent = additional_user_agent or []
         self._determined_user_agent = None
+        if discovery_cache is None:
+            discovery_cache = {}
+        self._discovery_cache = discovery_cache
+        # NOTE(mordred) split_loggers kwarg default is None rather than False
+        # so we can distinguish between the value being set or not.
+        self._split_loggers = split_loggers
 
         if timeout is not None:
             self.timeout = float(timeout)
@@ -298,17 +327,44 @@ class Session(object):
             return (header[0], '{SHA1}%s' % token_hash)
         return header
 
-    @positional()
+    def _get_split_loggers(self, split_loggers):
+        """Get a boolean value from the various argument sources.
+
+        We default split_loggers to None in the kwargs of the Session
+        constructor so we can track set vs. not set. We also accept
+        split_loggers as a parameter in a few other places. In each place
+        we want the parameter, if given by the user, to win.
+        """
+        # None is the default value in each method's kwarg. None means "unset".
+        if split_loggers is None:
+            # If no value was given, try the value set on the instance.
+            split_loggers = self._split_loggers
+        if split_loggers is None:
+            # If neither a value was given on the method, nor a value was given
+            # on the Session constructor, then the value is False.
+            split_loggers = False
+        return split_loggers
+
     def _http_log_request(self, url, method=None, data=None,
                           json=None, headers=None, query_params=None,
-                          logger=_logger):
+                          logger=None, split_loggers=None):
+        string_parts = []
+
+        if self._get_split_loggers(split_loggers):
+            logger = utils.get_logger(__name__ + '.request')
+        else:
+            # Only a single logger was passed in, use string prefixing.
+            string_parts.append('REQ:')
+            if not logger:
+                logger = utils.get_logger(__name__)
+
         if not logger.isEnabledFor(logging.DEBUG):
             # NOTE(morganfainberg): This whole debug section is expensive,
             # there is no need to do the work if we're not going to emit a
             # debug log.
             return
 
-        string_parts = ['REQ: curl -g -i']
+        string_parts.append('curl -g -i')
 
         # NOTE(jamielennox): None means let requests do its default validation
         # so we need to actually check that this is False.
@@ -331,7 +387,8 @@ class Session(object):
             string_parts.append(url)
 
         if headers:
-            for header in headers.items():
+            # Sort headers so that testing can work consistently.
+            for header in sorted(headers.items()):
                 string_parts.append('-H "%s: %s"'
                                     % self._process_header(header))
         if json:
@@ -346,10 +403,20 @@ class Session(object):
 
         logger.debug(' '.join(string_parts))
 
-    @positional()
     def _http_log_response(self, response=None, json=None,
                            status_code=None, headers=None, text=None,
-                           logger=_logger):
+                           logger=None, split_loggers=True):
+        string_parts = []
+        body_parts = []
+        if self._get_split_loggers(split_loggers):
+            logger = utils.get_logger(__name__ + '.response')
+            body_logger = utils.get_logger(__name__ + '.body')
+        else:
+            # Only a single logger was passed in, use string prefixing.
+            string_parts.append('RESP:')
+            body_parts.append('RESP BODY:')
+            body_logger = logger
+
         if not logger.isEnabledFor(logging.DEBUG):
             return
 
@@ -358,6 +425,19 @@ class Session(object):
                 status_code = response.status_code
             if not headers:
                 headers = response.headers
+
+        if status_code:
+            string_parts.append('[%s]' % status_code)
+        if headers:
+            # Sort headers so that testing can work consistently.
+            for header in sorted(headers.items()):
+                string_parts.append('%s: %s' % self._process_header(header))
+        logger.debug(' '.join(string_parts))
+
+        if not body_logger.isEnabledFor(logging.DEBUG):
+            return
+
+        if response is not None:
             if not text:
                 # NOTE(samueldmq): If the response does not provide enough info
                 # about the content type to decide whether it is useful and
@@ -382,25 +462,61 @@ class Session(object):
         if json:
             text = self._json.encode(json)
 
-        string_parts = ['RESP:']
-
-        if status_code:
-            string_parts.append('[%s]' % status_code)
-        if headers:
-            for header in headers.items():
-                string_parts.append('%s: %s' % self._process_header(header))
         if text:
-            string_parts.append('\nRESP BODY: %s\n' % text)
+            body_parts.append(text)
+            body_logger.debug(' '.join(body_parts))
 
-        logger.debug(' '.join(string_parts))
+    @staticmethod
+    def _set_microversion_headers(
+            headers, microversion, service_type, endpoint_filter):
+        # We're converting it to normalized version number for two reasons.
+        # First, to validate it's a real version number. Second, so that in
+        # the future we can pre-validate that it is within the range of
+        # available microversions before we send the request.
+        # TODO(mordred) Validate when we get the response back that
+        # the server executed in the microversion we expected.
+        # TODO(mordred) Validate that the requested microversion works
+        # with the microversion range we found in discovery.
+        microversion = discover.normalize_version_number(microversion)
+        # Can't specify a M.latest microversion
+        if (microversion[0] != discover.LATEST and
+                discover.LATEST in microversion[1:]):
+            raise TypeError(
+                "Specifying a '{major}.latest' microversion is not allowed.")
+        microversion = discover.version_to_string(microversion)
+        if not service_type:
+            if endpoint_filter and 'service_type' in endpoint_filter:
+                service_type = endpoint_filter['service_type']
+            else:
+                raise TypeError(
+                    "microversion {microversion} was requested but no"
+                    " service_type information is available. Either provide a"
+                    " service_type in endpoint_filter or pass"
+                    " microversion_service_type as an argument.".format(
+                        microversion=microversion))
+
+        # TODO(mordred) cinder uses volume in its microversion header. This
+        # logic should be handled in the future by os-service-types but for
+        # now hard-code for cinder.
+        if (service_type.startswith('volume') or
+                service_type == 'block-storage'):
+            service_type = 'volume'
+        headers.setdefault('OpenStack-API-Version',
+                           '{service_type} {microversion}'.format(
+                               service_type=service_type,
+                               microversion=microversion))
+        header_names = _mv_legacy_headers_for_service(service_type)
+        for h in header_names:
+            headers.setdefault(h, microversion)
 
-    @positional()
     def request(self, url, method, json=None, original_ip=None,
                 user_agent=None, redirect=None, authenticated=None,
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
-                endpoint_override=None, connect_retries=0, logger=_logger,
-                allow={}, client_name=None, client_version=None, **kwargs):
+                endpoint_override=None, connect_retries=0, logger=None,
+                allow=None, client_name=None, client_version=None,
+                microversion=None, microversion_service_type=None,
+                **kwargs):
         """Send an HTTP request with the specified characteristics.
 
         Wrapper around `requests.Session.request` to handle tasks such as
@@ -468,6 +584,16 @@ class Session(object):
         :type logger: logging.Logger
         :param dict allow: Extra filters to pass when discovering API
                            versions. (optional)
+        :param microversion: Microversion to send for this request.
+                       microversion can be given as a string or a tuple.
+                       (optional)
+        :param str microversion_service_type: The service_type to be sent in
+                       the microversion header, if a microversion is given.
+                       Defaults to the value of service_type from
+                       endpoint_filter if one exists. If endpoint_filter is not
+                       provided or does not have a service_type, microversion
+                       is given and microversion_service_type is not provided,
+                       an exception will be raised.
         :param kwargs: any other parameter that can be passed to
                        :meth:`requests.Session.request` (such as `headers`).
                        Except:
@@ -482,7 +608,19 @@ class Session(object):
 
         :returns: The response to the request.
         """
+        # If a logger is passed in, use it and do not log requests, responses
+        # and bodies separately.
+        if logger:
+            split_loggers = False
+        else:
+            split_loggers = None
+        logger = logger or utils.get_logger(__name__)
+
         headers = kwargs.setdefault('headers', dict())
+        if microversion:
+            self._set_microversion_headers(
+                headers, microversion, microversion_service_type,
+                endpoint_filter)
 
         if authenticated is None:
             authenticated = bool(auth or self.auth)
@@ -541,14 +679,14 @@ class Session(object):
             elif self.app_name:
                 agent.append(self.app_name)
 
-            for additional in self.additional_user_agent:
-                agent.append('%s/%s' % additional)
-
             if client_name and client_version:
                 agent.append('%s/%s' % (client_name, client_version))
             elif client_name:
                 agent.append(client_name)
 
+            for additional in self.additional_user_agent:
+                agent.append('%s/%s' % additional)
+
             if not agent:
                 # NOTE(jamielennox): determine_user_agent will return an empty
                 # string on failure so checking for None will ensure it is only
@@ -589,7 +727,7 @@ class Session(object):
                                    data=kwargs.get('data'),
                                    headers=headers,
                                    query_params=query_params,
-                                   logger=logger)
+                                   logger=logger, split_loggers=split_loggers)
 
         # Force disable requests redirect handling. We will manage this below.
         kwargs['allow_redirects'] = False
@@ -599,7 +737,7 @@ class Session(object):
 
         send = functools.partial(self._send_request,
                                  url, method, redirect, log, logger,
-                                 connect_retries)
+                                 split_loggers, connect_retries)
 
         try:
             connection_params = self.get_auth_connection_params(auth=auth)
@@ -631,13 +769,29 @@ class Session(object):
             request_id = (resp.headers.get('x-openstack-request-id') or
                           resp.headers.get('x-compute-request-id'))
             if request_id:
-                logger.debug('%(method)s call to %(service_name)s for '
-                             '%(url)s used request id '
-                             '%(response_request_id)s',
-                             {'method': resp.request.method,
-                              'service_name': service_name,
-                              'url': resp.url,
-                              'response_request_id': request_id})
+                if self._get_split_loggers(split_loggers):
+                    id_logger = utils.get_logger(__name__ + '.request-id')
+                else:
+                    id_logger = logger
+                if service_name:
+                    id_logger.debug(
+                        '%(method)s call to %(service_name)s for '
+                        '%(url)s used request id '
+                        '%(response_request_id)s', {
+                            'method': resp.request.method,
+                            'service_name': service_name,
+                            'url': resp.url,
+                            'response_request_id': request_id
+                        })
+                else:
+                    id_logger.debug(
+                        '%(method)s call to '
+                        '%(url)s used request id '
+                        '%(response_request_id)s', {
+                            'method': resp.request.method,
+                            'url': resp.url,
+                            'response_request_id': request_id
+                        })
 
         # handle getting a 401 Unauthorized response by invalidating the plugin
         # and then retrying the request. This is only tried once.
@@ -656,7 +810,7 @@ class Session(object):
 
         return resp
 
-    def _send_request(self, url, method, redirect, log, logger,
+    def _send_request(self, url, method, redirect, log, logger, split_loggers,
                       connect_retries, connect_retry_delay=0.5, **kwargs):
         # NOTE(jamielennox): We handle redirection manually because the
         # requests lib follows some browser patterns where it will redirect
@@ -702,13 +856,15 @@ class Session(object):
             time.sleep(connect_retry_delay)
 
             return self._send_request(
-                url, method, redirect, log, logger,
+                url, method, redirect, log, logger, split_loggers,
                 connect_retries=connect_retries - 1,
                 connect_retry_delay=connect_retry_delay * 2,
                 **kwargs)
 
         if log:
-            self._http_log_response(response=resp, logger=logger)
+            self._http_log_response(
+                response=resp, logger=logger,
+                split_loggers=split_loggers)
 
         if resp.status_code in self._REDIRECT_STATUSES:
             # be careful here in python True == 1 and False == 0
@@ -730,7 +886,7 @@ class Session(object):
                 # NOTE(jamielennox): We don't pass through connect_retry_delay.
                 # This request actually worked so we can reset the delay count.
                 new_resp = self._send_request(
-                    location, method, redirect, log, logger,
+                    location, method, redirect, log, logger, split_loggers,
                     connect_retries=connect_retries,
                     **kwargs)
 
@@ -852,9 +1008,46 @@ class Session(object):
         :returns: An endpoint if available or None.
         :rtype: string
         """
+        if 'endpoint_override' in kwargs:
+            return kwargs['endpoint_override']
+
         auth = self._auth_required(auth, 'determine endpoint URL')
+
         return auth.get_endpoint(self, **kwargs)
 
+    def get_endpoint_data(self, auth=None, **kwargs):
+        """Get endpoint data as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin:
+            if a plugin is not available.
+        :raises TypeError: If arguments are invalid
+
+        :returns: Endpoint data if available or None.
+        :rtype: keystoneauth1.discover.EndpointData
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_endpoint_data(self, **kwargs)
+
+    def get_api_major_version(self, auth=None, **kwargs):
+        """Get the major API version as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
+            plugin is not available.
+
+        :return: The major version of the API of the service discovered.
+        :rtype: tuple or None
+        """
+        auth = self._auth_required(auth, 'determine endpoint URL')
+        return auth.get_api_major_version(self, **kwargs)
+
     def get_auth_connection_params(self, auth=None, **kwargs):
         """Return auth connection params as provided by the auth plugin.
 
@@ -964,7 +1157,7 @@ class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
 
     This Adapter also preserves the default behaviour of Requests which
     disables Nagle's Algorithm. See also:
-    http://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
+    https://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
     """
 
     def init_poolmanager(self, *args, **kwargs):
